/* cbrp.cc
   */

extern "C" {
#include <stdarg.h>
#include <float.h>
};

#include "cbrpagent.h"
#include "priqueue.h"

#include <random.h>

#include <cmu/cmu-trace.h>
#include <cmu/marshall.h>
#include <packet.h>
#include <cmu/mac-802_11.h>
#define CBRP_ALMOST_NOW     0.1 // jitter used for events that should be effectively
				// instantaneous but are jittered to prevent
				// synchronization
#define CBRP_MIN_TUP_PERIOD 1.0 // minimum time between triggered updates
#define CBRP_IP_DEF_TTL   32 // default TTTL
#define MAX_LOCAL_REPAIR_TIMES 5

static const int verbose_srr = 0;

//the following codes are from DSR's implementation -Jinyang
Time cbrp_arp_timeout = 30.0e-3;     // (sec) arp request timeout
Time cbrp_rt_rq_period = 0.5;        // (sec) length of one backoff period
Time cbrp_rt_rq_max_period = 10.0;   // (sec) maximum time between rt reqs
Time cbrp_max_err_hold = 1.0;        // (sec)

/*************** selectors ******************/

bool cbrp_shortening_route_if_possible = true;
bool cbrp_local_repair = true;
bool cbrp_snoop_forwarded_errors = true; //same as DSR's snoop error

//with CBRP, you do get much benefit from snooping other nodes' source
//routes as DSR, netherless, we include this for people to try out themselves
bool cbrp_snoop_source_routes = false;

// should we only respond to the first route request we receive from a host?
bool cbrp_reply_only_to_first_rtreq = false;

// should we take the data from the last route error msg sent to us
// and propagate it around on the next propagating route request we do?
// this is aka grat route error propagation DSR uses this  
// to check invalid entries in route cache. As most of route replies in DSr
// comes from route cache, keeping route cache as updated as possible is 
// crutial to DSR's performance. In CBRP, most route replies comes from 
//destination node itself and hence does not depend too much on route cache.
bool cbrp_propagate_last_error = false;

// Should CBRP save the packet if the Local Repair procedure fails?
bool cbrp_salvage_with_cache = true;

// should we listen to a promiscuous tap?
// Unlike DSR, most of CBRP's RREP comes from destination node instead of
// cache and hence doing eavesdropping does no significant improvements.
bool cbrp_use_tap = false;

// CBRP does not use cache to reply when progagating RREQ.
bool cbrp_reply_from_cache_on_propagating = false;

// Used only when a node is eavesdropping.  
bool cbrpagent_reply_from_cache_uninvited = false;

// Returns a random number between 0 and max, from dmaltz
static inline double 
jitter (double max, int be_random_)
{
  return (be_random_ ? Random::uniform(max) : 0);
}

#ifdef CBRP_DEBUG
static char*
return_reply_path(Packet *p, int off_cbrp_)
{
  hdr_cbrp *cbrph = (hdr_cbrp *)p->access(off_cbrp_);
  static char buf[100];
  char *ptr = buf;
  char *rtn_buf = ptr;
  int len = cbrph->route_reply_len();

  if (len== 0)
    {
      sprintf(rtn_buf,"[<empty path>]");
      return rtn_buf;
    }
  *ptr++ = '[';
  for (int c = 0 ; c < len ; c ++)
    {
      ptr += sprintf(ptr,"%d ",cbrph->reply_addrs()[c].addr);
    }
  *ptr++ = ']';
  *ptr++ = '\0';
  return rtn_buf;   
}

static char*
return_RREQ_forward_nodes(Packet *p, int off_cbrp_)
{
  hdr_cbrp *cbrph = (hdr_cbrp *)p->access(off_cbrp_);
  static char buf[100];
  char *ptr = buf;
  char *rtn_buf = ptr;
  int len = cbrph->num_forwarders();

  if (len== 0)
    {
      sprintf(rtn_buf,"[<empty forwarders>]");
      return rtn_buf;
    }
  *ptr++ = '[';
  for (int c = 0 ; c < len ;c+=2)
    {
      ptr += sprintf(ptr,"%d(%d) ",cbrph->forwarders[c].which_head,cbrph->forwarders[c].which_gateway);
    }
  *ptr++ = ']';
  *ptr++ = '\0';
  return rtn_buf;   
}
#endif //CBRP_DEBUG

//the below two functions are from DSR's implementation -Jinyang :)
void CBRP_Agent:: trace (char *fmt,...)
{
  va_list ap;

  if (!logtarget)
    return;

  va_start (ap, fmt);
  vsprintf (logtarget->buffer (), fmt, ap);
  logtarget->dump ();
  va_end (ap);
}

void 
CBRP_Agent::tracepkt (Packet * p, double now, int me, const char *type)
{
  char buf[1024];

  unsigned char *walk = p->accessdata ();

  int ct = *(walk++);
  int seq, dst, met;

  snprintf (buf, 1024, "V%s %.5f _%d_ [%d]:", type, now, me, ct);
  while (ct--)
    {
      dst = *(walk++);
      met = *(walk++);
      seq = *(walk++);
      seq = seq << 8 | *(walk++);
      seq = seq << 8 | *(walk++);
      seq = seq << 8 | *(walk++);
      snprintf (buf, 1024, "%s (%d,%d,%d)", buf, dst, met, seq);
    }
  // Now do trigger handling.
  //trace("VTU %.5f %d", now, me);
  if (verbose_)
    trace ("%s", buf);
}

void
CBRP_Agent::recv(Packet * p, Handler *)
{
  hdr_ip *iph = (hdr_ip *) p->access (off_ip_);
  hdr_cmn *cmh = (hdr_cmn *) p->access (off_cmn_);
  hdr_cbrp *cbrph =  (hdr_cbrp*)p->access(off_cbrp_);

  /*
   *  Must be a packet I'm originating...
   */
  if(iph->src_ == myaddr_ && cmh->num_forwards() == 0) {
    iph->ttl_ = CBRP_IP_DEF_TTL;
  }
  /*
   *  Packet I'm forwarding...
   */
  else {
    /*
     *  Check the TTL.  If it is zero, then discard.
     */
    if(--iph->ttl_ == 0) {
      drop(p, DROP_RTR_TTL);
      return;
    }
  }
  


  if ((iph->src_ != myaddr_) &&  (!cbrph->valid()))
    {
      // process a BROADCAST pkt..
      ntable->processUpdate(p);
    }
  else
    {
      // process a CBRP source route pkt, RREP, RREQ, ERR pkt.
      forwardSRPacket(p);
      
    }
}

void
CBRP_Agent::tap(const Packet *packet)
/* CBRP does not need promiscuous mode, this is just for interested
   party to test out 
   The implementation is mostly taken from DSR -Jinyang*/
{

  hdr_cbrp *cbrph =  (hdr_cbrp*) ((Packet *)packet)->access(off_cbrp_);
  hdr_ip *iph =  (hdr_ip*)((Packet *)packet)->access(off_ip_);
  hdr_cmn *cmh =  (hdr_cmn*)((Packet *)packet)->access(off_cmn_);

  if (!cbrp_use_tap) return;

  if (!cbrph->valid()) return;    // can't do anything with it

  // don't trouble me with packets I'm about to receive anyway
  /* this change added 5/13/98 -dam */
  ID next_hop(cbrph->addrs[cbrph->cur_addr()].addr, ::IP);
  if (next_hop == net_id || next_hop == MAC_id) return;

  CBRP_Packet p((Packet *) packet, cbrph);
  p.dest = ID(iph->dst(),::IP);
  p.src = ID(iph->src(),::IP);

  if (p.src == net_id) return;

  /* snoop on the errori */
  if (cbrph->route_error())
    {
      if (verbose_)
        trace("Sdebug _%s_ tap saw error %d",  net_id.dump(), cmh->uid());
      processBrokenRouteError(p);
    }

  if (cbrph->route_reply())
    {
      Path reply_path(cbrph->reply_addrs(), cbrph->route_reply_len());
      reply_path.reverseInPlace();

      if(verbose_)
        trace("Sdebug _%s_ tap saw route reply %d  %s",
               net_id.dump(), cmh->uid(), reply_path.dump());
      route_cache->noticeRouteUsed(reply_path, Scheduler::instance().clock(),
                                   p.src);
      return;
    }

  if (cbrph->route_request() && ((unsigned long)iph->dst()) != IP_BROADCAST ) {
    if (cbrpagent_reply_from_cache_uninvited) {
      Packet *p_copy = p.pkt->copy();
      p.pkt = p_copy;
      //this copying is because replyFromRouteCache will free the pkt which we dont want
      replyFromRouteCache(p); 
    }
  }else if (cbrph->route_request() && ((unsigned long)iph->dst()) == IP_BROADCAST) {
    return;
  }

  // the logic is wrong for shortening rtreq's anyway, cur_addr always = 0

  if (cbrp_snoop_source_routes)
    {
      if (verbose_)
        trace("Sdebug _%s_ tap saw route use %d %s", net_id.dump(),
              cmh->uid(), p.route.dump());
      route_cache->noticeRouteUsed(p.route, Scheduler::instance().clock(), net_id);
    }

}

static class CBRPClass:public TclClass
{
  public:
  CBRPClass ():TclClass ("Agent/CBRP")
  {
  }
  TclObject *create (int, const char *const *)
  {
    return (new CBRP_Agent ());
  }
} class_cbrp;

CBRP_Agent::CBRP_Agent (): Agent (PT_CBRP), ll_queue (0),  
  myaddr_ (0), be_random_ (1), 
  use_mac_ (0), verbose_ (1), trace_wst_ (0), send_buf_timer(this),request_table(128),route_cache(NULL)
 {

  ntable = new NeighborTable(this);
  route_cache = makeRouteCache();

  route_request_num = 1;
  
  target_ = 0;
  logtarget = 0;
  
  bind("off_CBRP_", &off_cbrp_);
  bind("off_ll_", &off_ll_);
  bind("off_mac_", &off_mac_);
  bind("off_ip_", &off_ip_);

  route_error_held = false;

  bind ("use_mac_", &use_mac_);
  bind ("be_random_", &be_random_);
  bind ("verbose_", &verbose_);
  bind ("trace_wst_", &trace_wst_);
  bind ("no_of_clusters_", &no_of_clusters_); 
}

void
CBRP_Agent::startUp()
{
  // kick off periodic advertisments
  ntable->startUp();
}

/* This command function is mainly taken from DSR implementation */
int 
CBRP_Agent::command (int argc, const char *const *argv)
{
  TclObject *obj;

  if (argc == 2)
    {
      if (strcmp (argv[1], "startcbrp") == 0)
	{
	  startUp();
          send_buf_timer.sched(BUFFER_CHECK
                               + BUFFER_CHECK * Random::uniform(1.0));
          return (TCL_OK);
	}
      else if (strcmp (argv[1], "testinit") == 0) 
        {
          testinit();
          return (TCL_OK);
        }
      else if (strcmp (argv[1], "reset") == 0) 
        {
          Terminate();
          return Agent::command(argc, argv);
        }      
      else if (strcasecmp(argv[1], "check-cache") == 0)
        {
          return route_cache->command(argc, argv);
        }
      else if (strcasecmp (argv[1], "ll-queue") == 0)
	{
	if (!(ll_queue = (PriQueue *) TclObject::lookup (argv[2])))
	    {
	      fprintf (stderr, "CBRP_Agent: ll-queue lookup of %s failed\n", argv[2]);
	      return TCL_ERROR;
	    }

	  return TCL_OK;
	}

    }
  else if (argc == 3)
    {
      if (strcasecmp(argv[1], "ip-addr") == 0)
        {
         net_id = ID(atoi(argv[2]), ::IP);
         myaddr_ = net_id.addr; //myadd_ is the IP addr shortform
         route_cache->net_id = net_id;
         return TCL_OK;
        }
      else if(strcasecmp(argv[1], "mac-addr") == 0)
        {
          MAC_id = ID(atoi(argv[2]), ::MAC);
          route_cache->MAC_id = MAC_id;
          return TCL_OK;
        }

      if( (obj = TclObject::lookup(argv[2])) == 0)
        {
          fprintf(stderr, "CBRP_Agent: %s lookup of %s failed\n", argv[1],
                  argv[2]);
          return TCL_ERROR;
        }
      if (strcasecmp (argv[1], "log-target") == 0)  
        {
          logtarget = (Trace *)obj;
          return route_cache->command(argc, argv); 
        }
      else if (strcasecmp(argv[1], "install-tap") == 0)
        {
          Mac *m = (Mac*) obj;
          m->installTap(this);
          return TCL_OK;
        }

    }

  else if (argc == 4) 
    {
      if (strcasecmp(argv[1], "add-ll") == 0)
        {
          if( (obj = TclObject::lookup(argv[2])) == 0) {
            fprintf(stderr, "CBRP_Agent: %s lookup of %s failed\n", argv[1],
                    argv[2]);
            return TCL_ERROR;
          }
          ll = (NsObject*) obj;
          if( (obj = TclObject::lookup(argv[3])) == 0) {
            fprintf(stderr, "CBRP_Agent: %s lookup of %s failed\n", argv[1],
                    argv[3]);
            return TCL_ERROR;
          }
          ifq = (PriQueue *) obj;
          return TCL_OK;
        }
    }
return (Agent::command (argc, argv));
}

/*===========================================================================
  SendBuf management and helpers  -- from cmu/dsr/cbrp.cc by dmaltz
---------------------------------------------------------------------------*/
void
CBRP_SendBufferTimer::expire(Event *e) 
{ 
  a_->sendBufferCheck(); 
  resched(BUFFER_CHECK + BUFFER_CHECK * (double) ((int) e>>5 & 0xff) / 256.0);
}

void
CBRP_Agent::dropSendBuff(CBRP_Packet &p)
  // log p as being dropped by the sendbuffer in CBRP agent
{
  trace("Ssb %.5f _%s_ dropped %s -> %s", Scheduler::instance().clock(),
        net_id.dump(), p.src.dump(), p.dest.dump());
  drop(p.pkt, DROP_RTR_QTIMEOUT);
  p.pkt = 0;
  p.route.reset();
}

void
CBRP_Agent::stickPacketInSendBuffer(CBRP_Packet& p)
{
  Time min = DBL_MAX;
  int min_index = 0;
  int c;

  if (verbose_)
    trace("Sdebug %.5f _%s_ stuck into send buff %s -> %s",
          Scheduler::instance().clock(),
          net_id.dump(), p.src.dump(), p.dest.dump());

  for (c = 0 ; c < SEND_BUF_SIZE ; c ++)
    if (send_buf[c].p.pkt == NULL)
      {
        send_buf[c].t = Scheduler::instance().clock();
        send_buf[c].p = p;
        return;
      }
    else if (send_buf[c].t < min)
      {
        min = send_buf[c].t;
        min_index = c;
      }

  // kill somebody
  dropSendBuff(send_buf[min_index].p);
  send_buf[min_index].t = Scheduler::instance().clock();
  send_buf[min_index].p = p;
}

void
CBRP_Agent::sendBufferCheck()
  // see if any packets in send buffer need route requests sent out
  // for them, or need to be expired
{ // this is called about once a second.  run everybody through the
  // get route for pkt routine to see if it's time to do another
  // route request or what not
  int c;

  for (c  = 0 ; c <SEND_BUF_SIZE ; c++)
    {
      if (send_buf[c].p.pkt == NULL) continue;
      if (Scheduler::instance().clock() - send_buf[c].t > SEND_TIMEOUT)
        {
          dropSendBuff(send_buf[c].p);
          send_buf[c].p.pkt = 0;
          continue;
        }
#ifdef DEBUG
      trace("Sdebug %.5f _%s_ checking for route for dst %s",
            Scheduler::instance().clock(), net_id.dump(),
            send_buf[c].p.dest.dump());
#endif
      handlePktWithoutSR(send_buf[c].p, true);
#ifdef DEBUG
      if (send_buf[c].p.pkt == NULL)
        trace("Sdebug %.5f _%s_ sendbuf pkt to %s liberated by handlePktWOSR",
            Scheduler::instance().clock(), net_id.dump(),
            send_buf[c].p.dest.dump());
#endif
    }
}

/*==============================================================
  Route Request backoff -dmaltz
------------------------------------------------------------*/
static bool
BackOffTest(Entry *e, Time time)
// look at the entry and decide if we can send another route
// request or not.  update entry as well
{
  Time next = ((Time) (0x1<<(e->rt_reqs_outstanding*2))) * cbrp_rt_rq_period;
  if (next > cbrp_rt_rq_max_period) next = cbrp_rt_rq_max_period;
  if (next + e->last_rt_req > time) return false;
  // don't let rt_reqs_outstanding overflow next on the LogicalShiftsLeft's
  if (e->rt_reqs_outstanding < 15) e->rt_reqs_outstanding++;
  e->last_rt_req = time;
  return true;
}

void
CBRP_Agent::Terminate()
{
        int c;
        for (c  = 0 ; c < SEND_BUF_SIZE ; c++) {
                if (send_buf[c].p.pkt) {
                        drop(send_buf[c].p.pkt, DROP_END_OF_SIMULATION);
                        send_buf[c].p.pkt = 0;
                }
        }
}

void
CBRP_Agent::testinit()
{
  struct hdr_cbrp hsr;

  if (net_id == ID(1,::IP))
    {
      printf("adding route to 1\n");
      hsr.init();
      hsr.append_addr( 1, AF_INET );
      hsr.append_addr( 2, AF_INET );
      hsr.append_addr( 3, AF_INET );
      hsr.append_addr( 4, AF_INET );

      route_cache->addRoute(Path(hsr.addrs, hsr.num_addrs()), 0.0, ID(1,::IP));
    }

  if (net_id == ID(3,::IP))
    {
      printf("adding route to 3\n");
      hsr.init();
      hsr.append_addr( 3, AF_INET );
      hsr.append_addr( 2, AF_INET );

      route_cache->addRoute(Path(hsr.addrs, hsr.num_addrs()), 0.0, ID(3,::IP));
    }
}

/*-------------from here onwards, we deal with more CBRP specific stuff---*/

void 
CBRP_Agent::forwardSRPacket(Packet* packet)
{

  hdr_cbrp *cbrph =  (hdr_cbrp*)packet->access(off_cbrp_);
  hdr_ip *iph =  (hdr_ip*)packet->access(off_ip_);
  hdr_cmn *cmh =  (hdr_cmn*)packet->access(off_cmn_);

  assert(cmh->size() >= 0);

  CBRP_Packet p(packet, cbrph);
  p.dest = ID(iph->dst(),::IP);
  p.src = ID(iph->src(),::IP);

  assert(logtarget != 0);

  if (!cbrph->valid())
    {
      // this must be an outgoing packet, it doesn't have a SR header on it
      cbrph->init();               // give packet an SR header now
      if (verbose_)
        trace("S %.9f _%s_ originating %s -> %s",
              Scheduler::instance().clock(), net_id.dump(), p.src.dump(),
              p.dest.dump());
      cmh->size() += IP_HDR_LEN;
      cmh->ptype() = PT_CBR;
      handlePktWithoutSR(p, false);
      goto done;
    }
  
  if ( ((p.dest == net_id) && (!cbrph->route_request())) 
            || ((p.dest == IP_broadcast) && (!cbrph->route_request()))
            || (cbrph->route_request() && ((unsigned long)cbrph->request_destination() == net_id.addr)) )
    { 
      //this is either a route_reply for my previously sent-out RREQ or a data packet for me
      //or a route_request searching for me
      handlePacketReceipt(p);
      goto done;
    }

  // should we check to see if it's an error packet we're handling
  // and if so call processBrokenRouteError to snoop
  if (cbrp_snoop_forwarded_errors && cbrph->route_error())
    {
      processBrokenRouteError(p);
    }

  if (cbrph->route_request())
    { 
      // propagate a route_request that's not for us
      //if I am not a cluster head, forward the RREQ as indicated to p.dest
      //if I am cluster head, broadcast to other un-visited neighboring clusters
      handleRREQ(p);
    }
  else if (!cbrph->route_repaired() && !cbrph->route_shortened() && cbrph->route_reply())
    {
      //if I am cluster head, i have to calculate part of the RREP
      //if I am oridnary node, record myself in RREP and forward it to the next cluster head
      handleRREP(p);
    }
  else
    { // we're not the intended final receipient, but we're a hop along the route
      // since repaired route is hop by hop, we will use normal data packet forwarding
      // function to handle it
      handleForwarding(p);
    }

done:
  assert(p.pkt == 0);
  return;
}

/*--------------------------------------------------------------------------=
  different packet handling functions, 
---------------------------------------------------------------------------*/
void
CBRP_Agent::handlePktWithoutSR(CBRP_Packet& p, bool retry)
  /* obtain a source route to p's destination and send it off.
     this should be a retry if the packet is already in the sendbuffer */
{
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  hdr_ip *iph =  (hdr_ip*)p.pkt->access(off_ip_);
  hdr_cmn *ch = (hdr_cmn*)p.pkt->access(off_cmn_);

  assert(cbrph->valid());
  assert(iph->src() == myaddr_);
  if (p.dest == net_id)
    { 
      // it doesn't need a source route, 'cause it's for us
      handlePacketReceipt(p);
      return;
    }

  //firstly, try our route cache to see if the route has been discovered
  if (route_cache->findRoute(p.dest, p.route, 1))
    { 
      // lucky! we've got a route...
      if (verbose_)
        trace("S$hit %.5f _%s_ %s -> %s %s",
              Scheduler::instance().clock(), net_id.dump(),
              p.src.dump(), p.dest.dump(), p.route.dump());
      sendOutPacketWithRoute(p, true);
      return;
    } 

  //secondly, examine neighborhood table to see if destination node is our neighbor
  if (ntable->isNeighbor(p.dest.addr)) 
    {
      p.route.reset();
      p.route.appendToPath(net_id);
      p.route.appendToPath(p.dest);
      if (verbose_) {
        trace("S$direct hit %.5f _%s_ %d -> %s %s",
              Scheduler::instance().clock(), net_id.dump(),
              iph->src(), p.dest.dump(), p.route.dump());
      }
      sendOutPacketWithRoute(p, true);
      return;
    }

  
  //thirdly, we'll check using two-hop-neighbor table if we could reach
  // the destination using only one intermediate node
  if (ntable->GetQuickNextNode(p.dest.addr)) 
    {
      p.route.reset();
      p.route.appendToPath(net_id);
      p.route.appendToPath(ID(ntable->GetQuickNextNode(p.dest.addr),::IP));
      p.route.appendToPath(p.dest);
 
      if (verbose_) {
        trace("S$direct hit %.5f _%s_ %d -> %s %s",
              Scheduler::instance().clock(), net_id.dump(),
              iph->src(), p.dest.dump(), p.route.dump());
      }
      sendOutPacketWithRoute(p, true);
      return;
    }
 
    // We have to send out a RREQ to discover the destination 
     
    if (verbose_)
      trace("S$miss %.5f _%s_ %s -> %s retry %d",
            Scheduler::instance().clock(), net_id.dump(),
            net_id.dump(), p.dest.dump(), retry?1:0);
    getRouteForPacket(p, retry);

}

void
CBRP_Agent::handlePacketReceipt(CBRP_Packet& p)
  /* Handle a packet destined to us */
{
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);

  if (cbrph->route_reply())
    {
      //we've got a route_reply! 
      acceptRREP(p);
    }

  if (cbrph->route_request())
    {
      assert((unsigned long)cbrph->request_destination() == net_id.addr || net_id == p.dest);
      if (cbrp_reply_only_to_first_rtreq  && ignoreRouteRequestp(p))
        { //we only respond to the first route request
          // we receive from a host
          Packet::free(p.pkt);     // drop silently
          p.pkt = 0;
          return;
        }
      else
        { // we're going to process this request now, so record the req_num
          request_table.insert(p.src, p.src, cbrph->rtreq_seq());
          //initiate a RREP in reply to the RREQ
          returnSrcRouteToRequestor(p);
        }
    }

  if (cbrph->route_error())
    { // register the dead route
      processBrokenRouteError(p);
    }

  if (cbrph->route_shortening()) {
   //the previous route shortening attempt is successful, shorten the route
    for (int c = (cbrph->cur_addr()-1); c < cbrph->num_addrs()-1 ; c++)
    {
      p.route[c].addr = p.route[c+1].addr;
    }
    p.route.setLength(cbrph->num_addrs()-1);
    cbrph->route_shortened() = 1;
    cbrph->route_shortening() = 0;
  }

 
  if (!cbrph->route_reply() && (unsigned long)cbrph->addrs[0].addr == p.src.addr
      && (cbrph->route_shortened() || cbrph->route_repaired())) {
   
    CBRP_Packet p_copy = p;
    p_copy.pkt = allocpkt();
    p_copy.dest = p.src;
    p_copy.src = net_id;

    hdr_ip *new_iph =  (hdr_ip*)p_copy.pkt->access(off_ip_);
    hdr_cbrp *new_cbrph =  (hdr_cbrp*)p_copy.pkt->access(off_cbrp_);

    new_cbrph->init();
    new_cbrph->route_repaired() = cbrph->route_repaired();
    new_cbrph->route_shortened() = cbrph->route_shortened();
    new_cbrph->route_reply() = 1;
 
    new_iph->src() = net_id.addr;
    new_iph->dst() = p.src.addr;
    new_iph->dport() = RT_PORT;
    new_iph->sport() = RT_PORT;
    new_iph->ttl() = 255;


    p_copy.route.reverseInPlace();
    p_copy.route.resetIterator();
    fillCBRPPath(p_copy.route,new_cbrph);

    hdr_cmn *new_cmnh =  (hdr_cmn*)p_copy.pkt->access(off_cmn_);
    new_cmnh->ptype() = PT_CBRP;
    sendOutPacketWithRoute(p_copy,1);

  }

  target_->recv(p.pkt, (Handler*)0);
  p.pkt = 0;
}

void
CBRP_Agent::handleForwarding(CBRP_Packet &p)
  /* forward packet on to next host in source route,
   snooping as appropriate */
{
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  hdr_cmn *cmh =  (hdr_cmn*)p.pkt->access(off_cmn_);

  trace("cbrp %.9f _%s_ --- %d [%s -> %s] %s",
        Scheduler::instance().clock(), net_id.dump(), cmh->uid(),
        p.src.dump(), p.dest.dump(), cbrph->dump());

  // first make sure we are the ``current'' host along the source route.
  // if we're not, the previous node set up the source route incorrectly.
  assert(p.route[p.route.index()] == net_id || p.route[p.route.index()] == MAC_id);

  if (p.route.index() >= p.route.length())
    {
      fprintf(stderr,"dfu: ran off the end of a source route\n");
      trace("SDFU:  ran off the end of a source route\n");
      drop(p.pkt, DROP_RTR_ROUTE_LOOP);
      p.pkt = 0;
      // maybe we should send this packet back as an error...
      return;
    }

  // if there's a source route, maybe we should snoop it too, does not do it by default
  if (cbrp_snoop_source_routes)
    route_cache->noticeRouteUsed(p.route, Scheduler::instance().clock(),
                                 net_id);

  // sendOutPacketWithRoute will add in the size of the src hdr, so
  // we have to subtract it out here
  struct hdr_cmn *ch = HDR_CMN(p.pkt);

  if (cbrph->route_shortening()) {
    //the previous node has successfully shortened one node
    for (int c = (cbrph->cur_addr()-1); c < cbrph->num_addrs()-1 ; c++)
      {
         cbrph->addrs[c].addr = cbrph->addrs[c+1].addr;
         p.route[c].addr = p.route[c+1].addr;
      }
    cbrph->cur_addr() -= 1;
    p.route.setIterator(cbrph->cur_addr());
    cbrph->num_addrs()-= 1;
    p.route.setLength(cbrph->num_addrs());
    cbrph->route_shortening() = 0;
    cbrph->route_shortened() = 1;
    trace("CBRP %.5f _%d_ success-shorten %s", Scheduler::instance().clock(),myaddr_, p.route.dump());
  }
  ch->size() -= cbrph->size();
  sendOutPacketWithRoute(p, false);
}

int
CBRP_Agent::handleRREP(CBRP_Packet &p)
{
 
  hdr_cbrp *cbrph = (hdr_cbrp*)p.pkt->access(off_cbrp_);
  hdr_cmn *cmnh =  (hdr_cmn*)p.pkt->access(off_cmn_);

  cmnh->addr_type() = AF_INET;

  cmnh->xmit_failure_ = CBRP_XmitFailureCallback;
  cmnh->xmit_failure_data_ = (void *) this;

  //just a cautionary step...
  if (p.route.length()<=0) {

#ifdef CBRP_DEBUG
    printf("ERROR route len %d\n", p.route.length());
    printf("p.src %s, p.dest %s, me %s, route-reply %s\n",
            p.src.dump(),p.dest.dump(),net_id.dump(),return_reply_path(p.pkt, off_cbrp_));
#endif
    abort();
  }

  if ((cbrph->cur_addr() >= 0) && (p.route[p.route.index()] == net_id) && (ntable->my_status == CLUSTER_HEAD)) {
    // I am the cluster head and have to calculate the route
    cbrph->cur_addr()--;
 
    if (cbrph->cur_addr()<0) {
      //i am the first cluster head, should be able to reach src
      if (ntable->isNeighbor(p.dest.getNSAddr_t())) {
        cmnh->next_hop() = p.dest.getNSAddr_t();

      }else {

        if (!(cmnh->next_hop()=ntable->GetQuickNextNode(p.dest.getNSAddr_t()))) {
          trace("CBRP %.5f _%d_ drop-reply-no-route unable return route to %s",Scheduler::instance().clock(), myaddr_,p.dest.dump());
          Packet::free(p.pkt);
          p.pkt = NULL;
          return 0;
        }
      }

      //if there is a link between next hop and the previous calculated hop, i am bypassed.
      //otherwise include me in the calculated route. 
      //Right now, the head won't search in its two-hop-topology table 
      //for a non-cluster-head guy to take off its work as the current 2-hop topology table 
      //implementation is not complete.  A more efficient data structure
      //has to be implemented to speed up the search. :)
      if (!ntable->existsLink(cmnh->next_hop(), 
                             (cbrph->reply_addrs()[cbrph->route_reply_len()-1]).addr)) {
        (cbrph->reply_addrs()[cbrph->route_reply_len()]).addr = net_id.addr;
        (cbrph->reply_addrs()[cbrph->route_reply_len()]).addr_type = AF_INET;
        cbrph->route_reply_len()++;
      }
      cmnh->prev_hop_ = net_id.getNSAddr_t();

      cmnh->size() = cbrph->size();
      Scheduler::instance().schedule(ll, p.pkt, 0);
      goto done;
    }

    // I am not the first cluster head, I'll try to reach the prev. cluster head  
    nsaddr_t prev_cluster = (cbrph->addrs)[cbrph->cur_addr_].addr;
    nsaddr_t next_node = (cbrph->reply_addrs()[cbrph->route_reply_len()-1]).addr;

    // firstly try to reach the prev_cluster by 2 hops
    adjtable_ent *ent = ntable->GetAdjEntry(prev_cluster,1);

    if (ent == NULL) {
      //if i could not reach prev_cluster by 2 hops, try 3-hop gateway.
      ent = ntable->GetAdjEntry(prev_cluster, 0);
    }

    if (ent == NULL) {
      //I failed to reach prev_cluster
      trace("CBRP %.5f _%s_ drop-reply-no-route #%d (route reply %s -> %s cannot reach prev cluster %d on route %s) previous hop %d", Scheduler::instance().clock(), net_id.dump(), cbrph->rtreq_seq(), p.src.dump(), p.dest.dump(),prev_cluster, p.route.dump(), cmnh->prev_hop_);
      Packet::free(p.pkt);
      p.pkt = NULL;
      return 0;
    }

    nexthop_ent *next_hop = ent->next_hop;
    assert(next_hop != NULL);
  
    while (next_hop) {
      //try to search for a gateway node that could be directly reached by route's next node
      //therefore, bypassing me.
      if (ntable->existsLink(next_hop->next_node,next_node)) {
        cmnh->next_hop() = next_hop->next_node;
        cmnh->addr_type() = AF_INET;
        break;
      }
      next_hop = next_hop->next;
    }

    if (next_hop == NULL) {
      cmnh->next_hop() = ent->next_hop->next_node;
      assert(cmnh->next_hop()>0);
      // it has to go pass me. -Jinyang
      (cbrph->reply_addrs()[cbrph->route_reply_len()]).addr = net_id.addr;
      (cbrph->reply_addrs()[cbrph->route_reply_len()]).addr_type = AF_INET;
      cbrph->route_reply_len()++;
    }     
    cmnh->size() = cbrph->size();
    cmnh->prev_hop_ = net_id.getNSAddr_t();
    Scheduler::instance().schedule(ll, p.pkt, 0);

  } else { 

    //I am an ordinary node non-CLUSTER_HEAD

    nsaddr_t next_cluster;

    //if I am just an ordinary node, I just forward it to the next cluster head
    cmnh->addr_type() = AF_INET;

    if (cbrph->cur_addr_<0) {
      //I am supposed to be two hop away from the destination node
      next_cluster = p.dest.addr;
      
      if (ntable->isNeighbor(next_cluster)) {
        cmnh->next_hop() = next_cluster;
      }else {
        cmnh->next_hop() = ntable->GetQuickNextNode(next_cluster);
        if (!cmnh->next_hop()) {
          trace("CBRP %.5f _%d_ drop-reply-no-route %s->%s #%d (route reply cannot reach prev cluster %d)",Scheduler::instance().clock(), myaddr_,p.src.dump(), p.dest.dump(), cbrph->rtreq_seq(),next_cluster);
          Packet::free(p.pkt);
          p.pkt = NULL;
          return 0;
        }
      }
    }else {
      //try to reach the next_cluster head.
      next_cluster = cbrph->addrs[cbrph->cur_addr()].addr;
      //check if the neighbor cluster is just one hop away
      if (ntable->isNeighbor(next_cluster)) {
        cmnh->next_hop() = next_cluster;
      } else { 
        cmnh->next_hop() = ntable->GetQuickNextNode(next_cluster);
        if (!cmnh->next_hop()) {
          trace("CBRP %.5f _%d_ drop-reply-no-route %s->%s #%d (route reply cannot reach prev cluster %d)", Scheduler::instance().clock(), myaddr_,p.src.dump(), p.dest.dump(), cbrph->rtreq_seq(),next_cluster);
          Packet::free(p.pkt);
          p.pkt = NULL;
          return 0;
        }
      }
 
    }

    //checking for wrong route reply calculation which could cause looping.
    int i;
    for (i=cbrph->route_reply_len();i>0;i--) {
      if (cbrph->reply_addrs()[i-1].addr == cmnh->next_hop()) {
        break;
      }
    }
    if (i>0) {
      trace("CBRP %.5f _%d_ drop-reply-looped %s->%s #%d existing route:%s reaching next cluster %d via %d",Scheduler::instance().clock(),myaddr_, p.src.dump(), p.dest.dump(), cbrph->rtreq_seq(),p.route.dump(),next_cluster,cmnh->next_hop());
       Packet::free(p.pkt);
       p.pkt = NULL;
       return 0;
     }
 
    //add myself to the route reply
    (cbrph->reply_addrs()[cbrph->route_reply_len()]).addr = net_id.addr;
    (cbrph->reply_addrs()[cbrph->route_reply_len()]).addr_type = AF_INET;
    cbrph->route_reply_len()++;
    
    cmnh->prev_hop_ = net_id.getNSAddr_t();
    cmnh->size() = cbrph->size();

    //Scheduler::instance().schedule(ll, p.pkt, Random::uniform(RREQ_JITTER));
    Scheduler::instance().schedule(ll, p.pkt,0 );
  }

done:
#ifdef CBPR_DEBUG
  trace("cbrp _%d_ route-reply-sent %s #%d -> %s , next hop %d to reach %d, rest of recorded cluster heads %s, constructed return path %s", myaddr_, p.src.dump(),cbrph->rtreq_seq(), p.dest.dump(), cmnh->next_hop(),(cbrph->cur_addr()>-1)?cbrph->addrs[cbrph->cur_addr()].addr:iph->dst(),p.route.dump(),return_reply_path(p.pkt, off_cbrp_)); 
#endif
  p.pkt = NULL;
  return 1;
}


void
CBRP_Agent::handleRREQ(CBRP_Packet &p)
  /* process a route request that isn't targeted at us 
     cluster heads will broadcast RREQ to all chosen gateways
     Chosen gateways will unicast RREQ to neighbor cluster head*/
{
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  hdr_ip *iph = (hdr_ip*)p.pkt->access(off_ip_);

  assert (cbrph->route_request());
  if ((u_int32_t)iph->dst() != IP_BROADCAST) {
    //I am the neighborcluster head, i should broadcast this RREQ to my neighbors
    if (ntable->my_status == CLUSTER_HEAD || (iph->dst() == myaddr_)) {
       BroadcastRREQ(p);
    }else {
    //i should forward packet to the clusterhead as specified in ip destination
    //this could happen when the neighbor cluster head is 3 hops away and I am the second gateway
       UnicastRREQ(p,iph->dst());
    }

  }else {
   
    Packet *p_ori = p.pkt; 

    for (int i=0;i<cbrph->num_forwarders();i++) {


       if (cbrph->forwarders[i].which_gateway == myaddr_) {
          p.pkt = p_ori->copy();

          if ((cbrph->forwarders[i].which_head == myaddr_) && (ntable->my_status == CLUSTER_HEAD)) {
            //broadcast the pkt to other cluster heads
            BroadcastRREQ(p);
          }else {
            // if I am just an ordinary node, then just forward the RREQ as indicated in header
            UnicastRREQ(p,cbrph->forwarders[i].which_head);
          }
       }
    }
    // discard un-used RREQ packet
    if (p_ori) Packet::free(p_ori);

  }
  p.pkt = NULL;

}


void
CBRP_Agent::BroadcastRREQ(CBRP_Packet &p)
{

  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  hdr_cmn *cmnh = (hdr_cmn*)p.pkt->access(off_cmn_);
  hdr_ip *iph = (hdr_ip*)p.pkt->access(off_ip_);

  assert(cbrph->route_request());

  if (ignoreRouteRequestp(p))
    { 
      //-Jinyang only cluster heads records RREQs seen and drop RREQ if duplicate etc.
      if (verbose_srr)
        trace("cbrp %.5f _%s_ dropped %s #%d (ignored)",
            Scheduler::instance().clock(), net_id.dump(), p.src.dump(),
            cbrph->rtreq_seq());
      Packet::free(p.pkt);  // pkt is a route request we've already processed
      return; // drop silently

    }

  request_table.insert(p.src, p.src, cbrph->rtreq_seq());

  if (p.route.length() > (cbrph->max_propagation()/2))  
  // the chain of cluster heads are roughly twice as long as the hop by hop route
    {  
        if (verbose_srr)
          trace("cbrp %.5f _%s_ dropped %s #%d (prop limit exceeded)",
                Scheduler::instance().clock(), net_id.dump(), p.src.dump(),
                cbrph->rtreq_seq());
        Packet::free(p.pkt); // pkt isn't for us, and isn't data carrying
        p.pkt = 0;
        return;
     }


  if ((2*p.route.length()) > MAX_SR_LEN)
    {   
        // no propagation
        trace("cbrp %.5f _%s_ dropped %s #%d (SR full)",
              Scheduler::instance().clock(), net_id.dump(), p.src.dump(),
              cbrph->rtreq_seq());
        Packet::free(p.pkt);
        p.pkt = 0;
        return;
      }


  p.route.appendToPath(net_id);
  fillCBRPPath(p.route,cbrph);

  cmnh->ptype() = PT_CBRP;
  cmnh->addr_type() = AF_INET;

  if (ntable->isNeighbor(cbrph->request_destination())) {
      
    //directly unicast to the destination 
    cmnh->next_hop() = cbrph->request_destination();
    iph->dst() = cbrph->request_destination();
    cbrph->num_forwarders() = 0;
    cmnh->size() = cbrph->size();

    Scheduler &s = Scheduler::instance();
    s.schedule(ll,p.pkt,0.0);

    return;
       
   }

#ifdef CBRP_DEBUG
  trace("cbrp RREQ %d: %s",myaddr_,ntable->printNeighbors());
#endif
    
  if (ntable->GetQuickNextNode(cbrph->request_destination())) {

    iph->dst() = cbrph->request_destination();
    cmnh->next_hop() = ntable->GetQuickNextNode(cbrph->request_destination());

    cbrph->num_forwarders() = 0;
    cmnh->size() = cbrph->size();

    Scheduler::instance().schedule(ll,p.pkt,0.0);
    p.pkt = 0;
    return;
  }

  if (cbrp_reply_from_cache_on_propagating && replyFromRouteCache(p)) {
    p.pkt = NULL;
    return;
  }
 
  cmnh->next_hop() = MAC_BROADCAST;
  iph->dst() = IP_BROADCAST;

  int total = cbrph->num_forwarders();

  Packet *p_copy = p.pkt->copy();

  hdr_cbrp *new_cbrph = (hdr_cbrp*)p_copy->access(off_cbrp_);

  adjtable_ent *p_adj = ntable->adjtable_1;

  int i=0; 
  int j=0;

  while (p_adj && i<MAX_NEIGHBORS) {
    //check if a neighbor cluster head has been visited b4 by this RREQ by checking its recorded route so far
    if (p.route.member(ID(p_adj->neighbor_cluster, ::IP))) {

#ifdef CBRP_DEBUG
      trace("cbrp RREQ%d: ignored %d, duplicate in route(%s)",myaddr_,p_adj->neighbor_cluster,p.route.dump());
#endif
       p_adj = p_adj->next;
       continue;
     }
 
    //check if a neighbor cluster will have been visited by prev. cluster head by checking the forwarders' list
    for (j=0;j<total;j++) {

      if (p_adj->neighbor_cluster == cbrph->forwarders[j].which_head) {

#ifdef CBRP_DEBUG
        trace("cbrp RREQ%d: ignored %d, duplicate in previous broadcast",myaddr_,p_adj->neighbor_cluster);
#endif
        break;
      }
    }

    if (j>= total) {  //the next cluster head has NOT been visited to the best of our knowledge. :)

#ifdef CBRP_DEBUG
      trace("cbrp %.5f _%s_ rebroadcast %s #%d ->%d %s to 2 hop head %d via %d",Scheduler::instance().clock(), net_id.dump(), 
             p.src.dump(), cbrph->rtreq_seq(), cbrph->request_destination(), 
             p.route.dump(), p_adj->neighbor_cluster, 
             p_adj->next_hop->next_node);
#endif

      
      /*new_cbrph->reply_addrs()[i].which_gateway= p_adj->next_hop->next_node; */
      new_cbrph->forwarders[i].which_gateway= ntable->GetQuickNextNode(p_adj->neighbor_cluster);
      new_cbrph->forwarders[i++].which_head= p_adj->neighbor_cluster;

    }
    p_adj = p_adj->next;
  }  
 
  //if I am not a cluster head, i don't send RREQ to 3-hop neighbor cluster head 

  if (ntable->my_status != CLUSTER_HEAD) {
    goto DONE;
  }
   
  p_adj = ntable->adjtable_2;

  while (p_adj && i<MAX_NEIGHBORS) {

    if (p.route.member(ID(p_adj->neighbor_cluster, ::IP))) {

#ifdef CBRP_DEBUG
      trace("cbrp RREQ%d: ignored %d, route(%s)",myaddr_,p_adj->neighbor_cluster,p.route.dump());
#endif
       p_adj= p_adj->next;
       continue;
     }
    // i don't send to 3-hop neighbor head in cluster adjacency table if it is within 2-hop range
    if (ntable->GetAdjEntry(p_adj->neighbor_cluster,1)) {
#ifdef CBRP_DEBUG
      trace("cbrp RREQ%d: ignored (sent it to 2-hop gw) %d, route(%s)",myaddr_,p_adj->neighbor_cluster,p.route.dump());
#endif
      p_adj= p_adj->next;
      continue;
    }

    for (j=0;j<total;j++) {
      if (p_adj->neighbor_cluster == cbrph->forwarders[j].which_head) {
#ifdef CBRP_DEBUG
        trace("cbrp RREQ%d: ignored %d, duplicate in previous broadcast",myaddr_,p_adj->neighbor_cluster);
#endif
        break;
      }
    }
    if (j >= total ) {
      new_cbrph->forwarders[i].which_gateway= p_adj->next_hop->next_node;
      new_cbrph->forwarders[i++].which_head= p_adj->neighbor_cluster;
    }
    p_adj = p_adj->next;
  } 

DONE:
  new_cbrph->num_forwarders() = i;

  if (i != 0) {

    hdr_cmn *new_cmnh=(hdr_cmn*)p_copy->access(off_cmn_);
    new_cmnh->size() = new_cbrph->size();
#ifdef CBRP_DEBUG
    trace("cbrp %.5f _%d_ my route addres size %d forward path %s",Scheduler::instance().clock(), myaddr_, new_cbrph->num_addrs(),return_RREQ_forward_nodes(p_copy,off_cbrp_));
#endif
    trace("CBRP %.5f _%d_ broadcast-request %d->%d", Scheduler::instance().clock(),myaddr_, iph->src(),cbrph->request_destination());

    Scheduler::instance().schedule(ll,p_copy,0.0);
  }

  Packet::free(p.pkt);
  p.pkt = NULL;    

}


int 
CBRP_Agent::UnicastRREQ(CBRP_Packet &p, nsaddr_t nextcluster)
{
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  hdr_cmn *cmnh = (hdr_cmn*)p.pkt->access(off_cmn_);
  hdr_ip *iph = (hdr_ip*)p.pkt->access(off_ip_);

  cmnh->addr_type() = AF_INET;

  if (ntable->isNeighbor(cbrph->request_destination())) {
      
    //directly unicast to the destination 
    cmnh->next_hop() = cbrph->request_destination();
    iph->dst() = cbrph->request_destination();

    cbrph->num_forwarders() = 0;
    cmnh->size() = cbrph->size();

    Scheduler &s = Scheduler::instance();
    s.schedule(ll,p.pkt,0.0);
    return 1;
       
   }
  
  if (cbrp_reply_from_cache_on_propagating && replyFromRouteCache(p)) {
    p.pkt = NULL;
    return 1;
  }
 
  if (ntable->isNeighbor(nextcluster)) {

    cmnh->next_hop() = nextcluster;
    iph->dst() = nextcluster;

  }else if ((cmnh->next_hop() = ntable->GetQuickNextNode(nextcluster))) {

    iph->dst() = nextcluster;

  }else {

    trace("CBRP %.5f _%d_ drop-request : dropped(does not know how to reach next cluster %d), route(%s) cluster head? %s",Scheduler::instance().clock(),myaddr_,nextcluster,p.route.dump(), (ntable->my_status == CLUSTER_HEAD )? "yes":"no");

    Packet::free(p.pkt);
    p.pkt = NULL;
    return 0;
  }

  cmnh->xmit_failure_ = CBRP_XmitFailureCallback;
  cmnh->xmit_failure_data_ = (void *) this; 

  trace("cbrp %.5f _%s_ relay %s #%d ->%d %s to designated head %d via %d",Scheduler::instance().clock(), net_id.dump(), p.src.dump(), 
    cbrph->rtreq_seq(), cbrph->request_destination(), p.route.dump(),
    nextcluster, cmnh->next_hop());

  cmnh->size() = cbrph->size();

  Scheduler::instance().schedule(ll,p.pkt, 0.0);
  p.pkt = NULL;
  return 1;

}

/*===========================================================================
  Helpers
---------------------------------------------------------------------------*/
bool
CBRP_Agent::ignoreRouteRequestp(CBRP_Packet &p)
// should we ignore this route request? by dmaltz from DSR
{
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);

  if (request_table.get(p.src) >= cbrph->rtreq_seq())
    { // we've already processed a copy of this reqest so
      // we should drop the request silently
      return true;
    }
  if (p.route.member(net_id,MAC_id))
    { // we're already on the route, drop silently
      return true;
    }

  if (p.route.full())
    { 
      return true;
    }
  return false;
}


bool
CBRP_Agent::replyFromRouteCache(CBRP_Packet &p)
  /* - see if can reply to this route request from our cache
     if so, do it and return true, otherwise, return false
     - frees or hands off p iff returns true by dmaltz from DSR,modified for CBRP
     - CBRP does not reply from route-cache by default*/
{
  
  Path rest_of_route;

  // do we have a cached route the target?
  /* XXX what if we have more than 1?  (and one is legal for reply from
     cache and one isn't?) 1/28/97 -dam */
  hdr_cbrp *old_cbrph = (hdr_cbrp *)p.pkt->access(off_cbrp_);
  p.dest =  ID(old_cbrph->request_destination(),::IP);
  if (!route_cache->findRoute(p.dest, rest_of_route, 0))
    { // no route => we're done
      return false;
    }

  /* but we should be on on the remainder of the route (and should be at
   the start of the route */
  assert(rest_of_route[0] == net_id || rest_of_route[0] == MAC_id);

  if (rest_of_route.length() + 2 * p.route.length() > MAX_SR_LEN)
    return false; // too long to work with...

  // if there is any other information piggybacked into the
  // route request pkt, we need to forward it on to the dst
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  int request_seqnum = cbrph->rtreq_seq();

  // make up and send out a route reply
  Packet* rrp = allocpkt();

  hdr_ip *iph =  (hdr_ip*)rrp->access(off_ip_);
  iph->src() = myaddr_;
  iph->sport() = RT_PORT;
  iph->dst() = p.src.addr;
  iph->dport() = RT_PORT;
  iph->ttl() = 255;

  cbrph = (hdr_cbrp*)rrp->access(off_cbrp_);
  cbrph->init();
  cbrph->route_reply() = 1;

  int len = rest_of_route.length();
  for (int i = 0 ; i < len ; i++)
    rest_of_route[len - i - 1].fillSRAddr(cbrph->reply_addrs()[i]);
  cbrph->route_reply_len() = len; 

  // propagate the request sequence number in the reply for analysis purposes
  cbrph->rtreq_seq() = request_seqnum;

  hdr_cmn *cmnh =  (hdr_cmn*)rrp->access(off_cmn_);
  cmnh->ptype() = PT_CBRP;
  cmnh->size() = cbrph->size();

  // copy the rest of cluster head by cluster head route over
  fillCBRPPath(p.route,cbrph);
  cbrph->cur_addr() = cbrph->num_addrs() - 1;

  nsaddr_t next_cluster;
  if (cbrph->cur_addr() >= 0) {
    next_cluster = cbrph->addrs[cbrph->cur_addr()].addr;
  }else {
    next_cluster = iph->dst();
  }

  adjtable_ent *adj_ent;

  if ((cmnh->next_hop() = ntable->GetQuickNextNode(next_cluster))) {
    goto DONE;
  }else if ((adj_ent = ntable->GetAdjEntry(next_cluster, 0))) {
    cmnh->next_hop() = adj_ent->next_hop->next_node;
    goto DONE;
  }else {
    Packet::free(rrp);
    return false;
  }

DONE:
  Scheduler::instance().schedule(ll,rrp,0);     
  trace("CBRP %.9f _%s_ cache-reply-sent %d -> %d #%d (len %d) %s",
        Scheduler::instance().clock(), net_id.dump(),
        iph->src(), iph->dst(), request_seqnum, rest_of_route.length(),
        rest_of_route.dump());
  Packet::free(p.pkt);
  p.pkt = NULL;
  return true;
}

void
CBRP_Agent::sendOutPacketWithRoute(CBRP_Packet& p, bool fresh, Time delay = 0.0)
     // take packet and send it out, packet must a have a route in it
     // This packet should not not be a RREQ packet
     // if fresh is true then reset the path before using it, if fresh
     //  is false then our caller wants us use a path with the index
     //  set as it currently is
{


  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  hdr_cmn *cmnh = (hdr_cmn*)p.pkt->access(off_cmn_);

  assert(cbrph->valid());
  assert(!cbrph->route_request());

  if ((p.dest == net_id))  //Jinyang 
    { // it doesn't need to go on the wire, 'cause it's for us
      recv(p.pkt, (Handler *) 0);
      p.pkt = 0;
      return;
    }
  
  if (fresh)
    {
      p.route.resetIterator();
      if (verbose_)
        {
          trace("SO %.9f _%s_ originating %s, RREP? %s",
                Scheduler::instance().clock(),
                net_id.dump(), p.route.dump(), cbrph->route_reply()?"yes":"no" );
        }
    }
  fillCBRPPath(p.route,cbrph);
  assert((unsigned long)cbrph->addrs[cbrph->cur_addr()].addr == net_id.addr);

  // check if we could shorten the route by looking at our two-hop-topology table
  // we don't shorten the route for route errors and RREP (route reply)
  if ((!cbrph->route_error()) && (!cbrph->route_reply()) &&
       cbrp_shortening_route_if_possible && (cbrph->cur_addr()< cbrph->num_addrs()-2)) {

        //we only check to see if the node after next is reachable or not.
        //cause that's the most likely case. :)
        nsaddr_t next_next_node = cbrph->addrs[cbrph->cur_addr()+2].addr;

        if (ntable->isNeighbor(next_next_node)) {

          //I'll send the packet directly to the next next node. :)
          cmnh->next_hop() = next_next_node;
          cbrph->cur_addr() += 2;
          //set the flag to indicate I'm temporarily shortening the route, 
          //no changes are made to route at this point of time.
          cbrph->route_shortening() = 1; 
          trace("cbrp %.5f _%d_ trying to shorten the route %d->%d",Scheduler::instance().clock(),myaddr_,myaddr_,cbrph->addrs[cbrph->cur_addr()].addr);

        }else {
          cmnh->next_hop() = cbrph->get_next_addr();
          cbrph->cur_addr() = cbrph->cur_addr() + 1;
        }      
  } else {
        cmnh->next_hop() = cbrph->get_next_addr();
        cbrph->cur_addr() = cbrph->cur_addr() + 1;
  }
  cmnh->xmit_failure_ = CBRP_XmitFailureCallback;
  cmnh->xmit_failure_data_ = (void *) this;
  cmnh->prev_hop_ = net_id.getNSAddr_t();
  cmnh->addr_type() = AF_INET;
   
  assert(p.pkt->incoming == 0); // this is an outgoing packet

  if (ifq->length() > 25)
    trace("SIFQ %.5f _%s_ len %d",
          Scheduler::instance().clock(), net_id.dump(), ifq->length());

  //calculate the packet size
  if (cbrph->route_reply() || cbrph->route_error()) {
    cmnh->size() = cbrph->size();
  }else {
    cmnh->size() += cbrph->size();
  }

  // off it goes!!!!
  //assert(cmnh->ptype() == PT_CBR || cbrph->route_reply() || cbrph->route_error());
  Scheduler::instance().schedule(ll, p.pkt, delay);
  p.pkt = NULL; 


}

void
CBRP_Agent::getRouteForPacket(CBRP_Packet &p, bool retry)
  /* try to obtain a route for packet
     pkt is freed or handed off as needed, unless retry == true
     in which case it is not touched */
{

  Entry *e = request_table.getEntry(p.dest);
  Time time = Scheduler::instance().clock();

  if (!retry)
    {
      stickPacketInSendBuffer(p);
      p.pkt = 0; // pkt is handled for now (it's in sendbuffer)
    }

 
  /* make the route request packet */
  CBRP_Packet rrq;

  rrq.dest = ID(0, ::IP);  // the ip destination of RREQ is not necessarily same the destinate node
  rrq.src = net_id;

  rrq.pkt = allocpkt();
  hdr_cbrp *cbrph =  (hdr_cbrp*)rrq.pkt->access(off_cbrp_);
  hdr_ip *iph =  (hdr_ip*)rrq.pkt->access(off_ip_);
  hdr_cmn *cmnh =  (hdr_cmn*)rrq.pkt->access(off_cmn_);

  iph->dport() = RT_PORT;
  iph->src() = net_id.getNSAddr_t();
  iph->sport() = RT_PORT;

  cmnh->ptype() = PT_CBRP;
  cmnh->num_forwards() = 0;
  cmnh->addr_type() = AF_INET;

  cbrph->init();
  
  //set the requested destination node
  cbrph->route_request() = 1;
  cbrph->request_destination() = p.dest.getNSAddr_t();

  if (BackOffTest(e, time))
    { // it's time to start another route request cycle
      // CBRP does not have ring_zero search, but we still set the appropriate
      // e->last_type for reusing this part of DSR's code
      e->last_type = UNLIMIT;
      sendOutRtReq(rrq, MAX_SR_LEN);
      e->last_arp = time;

    }
  else if (LIMIT0 == e->last_type && (time - e->last_arp) > cbrp_arp_timeout)
   { 
     // try propagating rt req since we haven't heard back from limited one
     // Actually this part of the code will never be executed by CBRP - Li Jinyang
     e->last_type = UNLIMIT;
     sendOutRtReq(rrq, MAX_SR_LEN);
   }
  else
    { 
      // it's not time to send another route request...
      if (!retry && verbose_srr) {
        trace("SRR %.5f _%s_ RR-not-sent %s -> %s",
              Scheduler::instance().clock(),
              net_id.dump(), rrq.src.dump(), p.dest.dump());
      }
      trace("cbrp %s -> %s, request entry rt_reqs_outstanding %d, last_rt_req %.9f",rrq.src.dump(), p.dest.dump(),e->rt_reqs_outstanding, e->last_rt_req);

      Packet::free(rrq.pkt); // dump the route request packet we made up
      rrq.pkt = 0;
    }


}

void
CBRP_Agent::sendOutRtReq(CBRP_Packet &p, int max_prop)
  // Send out the Route request packet we have made
{

  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  hdr_ip *iph =  (hdr_ip*)p.pkt->access(off_ip_);
  hdr_cmn *cmnh = (hdr_cmn*)p.pkt->access(off_cmn_);
  int i=0;

  assert(cbrph->valid());

  cbrph->rtreq_seq() = route_request_num++;
  cbrph->max_propagation() = max_prop;

  p.route.reset();

  // CBRP does not propagate last error seen on Route Request, if you are
  // interested to see how this might help as it does for DSR, turn it on and try
  if (cbrp_propagate_last_error && route_error_held
      && Scheduler::instance().clock() - route_error_data_time  < cbrp_max_err_hold)
    {
      assert(cbrph->num_route_errors() < MAX_ROUTE_ERRORS);
      cbrph->route_error() = 1;
      link_down_cbrp *deadlink = &(cbrph->down_links()[cbrph->num_route_errors()]);
      deadlink->addr_type = AF_INET;
      deadlink->from_addr = err_from.getNSAddr_t();
      deadlink->to_addr = err_to.getNSAddr_t();
      deadlink->tell_addr = GRAT_ROUTE_ERROR;
      cbrph->num_route_errors() += 1;
      /*
       * Make sure that the Route Error gets on a propagating request.
       */
      if(max_prop > 0) route_error_held = false;
    }

  trace("CBRP %.5f _%s_ new-request %d %s #%d -> %d",
      Scheduler::instance().clock(), net_id.dump(),
      max_prop, p.src.dump(), cbrph->rtreq_seq(), cbrph->request_destination());

  //Jinyang - if i am the cluster head, append myself to the path before sending RREQ
  if (ntable->my_status == CLUSTER_HEAD) {

    p.route.appendToPath(net_id);
    fillCBRPPath(p.route,cbrph);

    //I will include all my 2-hop neighbor cluster head on forwarders list
    adjtable_ent *p_ent = ntable->adjtable_1;
    while (p_ent && (i<MAX_NEIGHBORS)) {
      cbrph->forwarders[i].which_gateway= p_ent->next_hop->next_node;
      cbrph->forwarders[i].which_head = p_ent->neighbor_cluster;

#ifdef CBRP_DEBUG
      trace("cbrp %.5f _%s_(head) RREQ(%d->%d) broadcasted to 2-hop head %d via %d", Scheduler::instance().clock(),net_id.dump(),
             cbrph->rtreq_seq(), cbrph->request_destination(),p_ent->neighbor_cluster,
             p_ent->next_hop->next_node);
#endif
      i++;
      p_ent = p_ent->next;
    }

    // I will include all my 3-hop neighbor cluster head on forwarders list
    p_ent = ntable->adjtable_2;

    while (p_ent && (i<MAX_NEIGHBORS)) {
      cbrph->forwarders[i].which_gateway= p_ent->next_hop->next_node;
      cbrph->forwarders[i].which_head= p_ent->neighbor_cluster;

#ifdef CBRP_DEBUG
      trace("cbrp %.5f _%s_(head) RREQ(%d->%d) broadcasted to 3-hop head %d via %d", Scheduler::instance().clock(),net_id.dump(),
               cbrph->rtreq_seq(), cbrph->request_destination(),p_ent->neighbor_cluster,
               p_ent->next_hop->next_node);
#endif
      i++;
      p_ent = p_ent->next;
    }
  }else {

    //I have to send the route request to one of my cluster head
    ntable_ent *ent = ntable->head;

    while (ent && (i< MAX_NEIGHBORS)) {
      if (ent->neighbor_status == CLUSTER_HEAD) {
        cbrph->forwarders[i].which_gateway= ent->neighbor;
        cbrph->forwarders[i].which_head = ent->neighbor;
#ifdef CBRP_DEBUG
        trace("cbrp %.5f _%s_ RREQ(%d->%d) broadcasted to 1-hop head %d",
               Scheduler::instance().clock(),net_id.dump(),cbrph->rtreq_seq(),
               cbrph->request_destination(),ent->neighbor);
#endif
        i++;
      }
      ent = ent->next;
    }
 
    // I will include all 2-hop cluster head in forwarders list 
    adjtable_ent *adjent=ntable->adjtable_1;

    while (adjent && (i<MAX_NEIGHBORS)) {

      // if the 2-hop cluster head is "really" not directly reachable
      if (!ntable->GetEntry(adjent->neighbor_cluster)) {  

         cbrph->forwarders[i].which_gateway= adjent->next_hop->next_node;
         cbrph->forwarders[i].which_head= adjent->neighbor_cluster;

#ifdef CBRP_DEBUG
         trace("cbrp %.5f _%s_ RREQ(%d->%d) broadcasted to 2-hop head %d via %d",
               Scheduler::instance().clock(),net_id.dump(),cbrph->rtreq_seq(),
               cbrph->request_destination(),adjent->neighbor_cluster,adjent->next_hop->next_node);
#endif
         i++;
     }
      adjent = adjent->next;
    }

   }
   cbrph->num_forwarders() = i;
   cmnh->next_hop() = MAC_BROADCAST;
   iph->dst() = IP_BROADCAST;
   iph->src() = myaddr_;

   cmnh->size() = cbrph->size();
   Scheduler::instance().schedule(ll,p.pkt,0.0);
}
   

void
CBRP_Agent::returnSrcRouteToRequestor(CBRP_Packet &p)
  // take the route in p, add us to the end of it and return the
  // RREP to the sender of p
  // doesn't free p.pkt
{

  hdr_cbrp *old_cbrph = (hdr_cbrp*)p.pkt->access(off_cbrp_);

  if (p.route.full())
    return; // alas, the route would be to long once we add ourselves

  CBRP_Packet p_copy = p;
  p_copy.pkt = allocpkt();
  p_copy.dest = p.src;

  hdr_ip *new_iph =  (hdr_ip*)p_copy.pkt->access(off_ip_);
  hdr_cbrp *new_cbrph =  (hdr_cbrp*)p_copy.pkt->access(off_cbrp_);
  hdr_cmn *new_cmnh =  (hdr_cmn*)p_copy.pkt->access(off_cmn_);

  new_cbrph->init();
  new_cbrph->route_reply() = 1;

  trace("CBRP %.5f _%s_ reply-sent %s->%d %s",
         Scheduler::instance().clock(),net_id.dump(),
         p.src.dump(),old_cbrph->request_destination(),p.route.dump());

  assert(old_cbrph->request_destination() == net_id.getNSAddr_t());

  p_copy.src = net_id;

  //if i am the cluster head, i have to complete the recorded route first
  if (ntable->my_status == CLUSTER_HEAD) {
   new_cbrph->reply_addrs()[0].addr = net_id.addr;
   new_cbrph->reply_addrs()[0].addr_type = AF_INET;
   new_cbrph->route_reply_len() = 1;
   p_copy.route.appendToPath(net_id);
  }
  fillCBRPPath(p_copy.route,new_cbrph);

  // just in case the route is empty
  if (!old_cbrph->num_addrs()) {
     //try to reach the sender directly!
     if ((new_cmnh->next_hop() = ntable->GetQuickNextNode(p_copy.dest.addr))) {

       new_cbrph->reply_addrs()[0].addr = p_copy.src.addr;
       new_cbrph->reply_addrs()[0].addr_type = AF_INET;
       new_cbrph->route_reply_len() = 1;
       new_iph->dst() = p_copy.dest.addr;
       new_iph->src() = myaddr_;
       new_cmnh->size() = new_cbrph->size();
       Scheduler::instance().schedule(ll,p_copy.pkt,0.0);

     }else {
       Packet::free(p_copy.pkt);
     }
     return;
   }

    //initialize route index pointer
  new_cbrph->cur_addr() = p_copy.route.length()-1;
  p_copy.route.setIterator(new_cbrph->cur_addr());


  new_iph->dst() = p_copy.dest.addr;
  new_iph->dport() = RT_PORT;
  new_iph->src() = p_copy.src.addr;
  new_iph->sport() = RT_PORT;
  new_iph->ttl() = 255;


  // propagate the request sequence number in the reply for analysis purposes
  new_cbrph->rtreq_seq() = old_cbrph->rtreq_seq();
  new_cmnh->ptype() = PT_CBRP;
  handleRREP(p_copy);

}

void
CBRP_Agent::acceptRREP(CBRP_Packet &p)
  /* - enter the packet's source route into our cache
     - see if any packets are waiting to be sent out with this source route
     - doesn't free the pkt */
{
  Path reply_route;
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);

  if (cbrph->route_repaired()) {
    trace("CBRP %.5f _%d_ grat-reply-repair from %d, route %s", Scheduler::instance().clock(),myaddr_, p.src.addr, p.route.dump());
  }else if (cbrph->route_shortened()){
    trace("CBRP %.5f _%d_ grat-reply-shorten from %d, route %s", Scheduler::instance().clock(),myaddr_, p.src.addr, p.route.dump());
  }else {
    trace("CBRP %.5f _%d_ reply-received from %d, route %s", Scheduler::instance().clock(),myaddr_,p.src.addr,p.route.dump());
  }

  if (!cbrph->route_reply())
    { // somethings wrong...
      trace("SDFU non route containing packet given to acceptRREP");
      fprintf(stderr, "dfu: non route containing packet given to acceptRRP\n");
    }

  if ((!cbrph->route_repaired()) && (!cbrph->route_shortened())) {
    Path reverse_reply_route(cbrph->reply_addrs(), cbrph->route_reply_len());
    //Jinyang - in CBRP, the returned source route is in reverse order
    reverse_reply_route.appendToPath(net_id);
    reply_route = reverse_reply_route.reverse();
  }else {
    Path reverse_reply_route(cbrph->addrs, cbrph->num_addrs());
    reply_route = reverse_reply_route.reverse();
  }

  /* check to see if this reply is valid or not using god info */
  int i;
  bool good_reply = true;
  for (i = 0; i < reply_route.length()-1 ; i++)
    if (God::instance()->hops(reply_route[i].getNSAddr_t(),
                              reply_route[i+1].getNSAddr_t()) != 1)
      {
        good_reply = false;
        break;
      }

#ifdef CBRP_DEBUG
  trace("cbrp %.9f _%s_ reply-received %d from %s  %s #%d -> %s %s",
        Scheduler::instance().clock(), net_id.dump(),
        good_reply ? 1 : 0,
        p.src.dump(), reply_route[0].dump(), cbrph->rtreq_seq(),
        reply_route[reply_route.length()-1].dump(),
        reply_route.dump());
#endif

  // add the new route into our cache
  route_cache->addRoute(reply_route, Scheduler::instance().clock(), p.src);

  // add the new loose source route to our loose source route cache
  Path source_route(cbrph->addrs,cbrph->num_addrs());

  // back down the route request counters
  Entry *e = request_table.getEntry(reply_route[reply_route.length()-1]);
  e->rt_reqs_outstanding = 0;
  e->last_rt_req = 0.0;

  // see if the addtion of this route allows us to send out
  // any of the packets we have waiting
  Time delay = 0.0;
  for (int c = 0; c < SEND_BUF_SIZE; c++)
    {
      if (send_buf[c].p.pkt == NULL) continue;
      if (route_cache->findRoute(send_buf[c].p.dest,
                                 send_buf[c].p.route, 1))
        { // we have a route!
#ifdef CBRP_DEBUG
          struct hdr_cmn *ch = HDR_CMN(send_buf[c].p.pkt);
          if(ch->size() < 0) {
                drop(send_buf[c].p.pkt, "XXX");
                abort();
          }
#endif

          if (verbose_)
            trace("Sdebug %.9f _%s_ liberated from sendbuf %s->%s %s",
                  Scheduler::instance().clock(), net_id.dump(),
                  send_buf[c].p.src.dump(), send_buf[c].p.dest.dump(),
                  send_buf[c].p.route.dump());
          
          sendOutPacketWithRoute(send_buf[c].p, true, delay);
          /* DSR spread out the rate in order to arp to complete, since
             we've already optimized ARP to take in MAC address of neighbors
             in advance, we don't really need to delay any more */
          //delay += cbrp_arp_timeout;
          send_buf[c].p.pkt = NULL;
        }
    }
}

void
CBRP_Agent::processBrokenRouteError(CBRP_Packet& p)
// take the error packet and proccess our part of it.
// if needed, send the remainder of the errors to the next person
// doesn't free p.pkt, mostly taken from DSR's implementation by dmaltz
{
  hdr_cbrp *cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);

  if (!cbrph->route_error())
    return; // what happened??

  /* if we hear A->B is dead, should we also run the link B->A through the
     cache as being dead, since 802.11 requires bidirectional links
      XXX -dam 4/23/98 */

  // since CPU time is cheaper than network time, we'll process
  // all the dead links in the error packet
  assert(cbrph->num_route_errors() > 0);
  int count = cbrph->num_route_errors();
  for (int c = 0 ; c < count ; c++)
    {
      assert(cbrph->down_links()[c].addr_type == AF_INET);
      route_cache->noticeDeadLink(ID(cbrph->down_links()[c].from_addr,::IP),
                                 ID(cbrph->down_links()[c].to_addr,::IP),
                                 Scheduler::instance().clock());
      // I'll assume everything's of type AF_INET for the printout... XXX
      if (verbose_srr)
        trace("SRR %.9f _%s_ dead-link tell %d  %d -> %d",
              Scheduler::instance().clock(), net_id.dump(),
              cbrph->down_links()[c].tell_addr,
              cbrph->down_links()[c].from_addr,
              cbrph->down_links()[c].to_addr);
    }

  ID who = ID(cbrph->down_links()[cbrph->num_route_errors()-1].tell_addr, ::IP);
  if (who != net_id && who != MAC_id)
    { // this error packet wasn't meant for us to deal with
      // since the outer entry doesn't list our name
      return;
    }

  // record this route error data for possible propagation on our next
  // route request
  route_error_held = true;
  err_from = ID(cbrph->down_links()[cbrph->num_route_errors()-1].from_addr,::IP);
  err_to = ID(cbrph->down_links()[cbrph->num_route_errors()-1].to_addr,::IP);
  route_error_data_time = Scheduler::instance().clock();

  if (1 == cbrph->num_route_errors())
    { // this error packet has done its job
      // it's either for us, in which case we've done what it sez
      // or it's not for us, in which case we still don't have to forward
      // it to whoever it is for
      return;
    }

  /* make a copy of the packet and send it to the next tell_addr on the
     error list.  the copy is needed in case there is other data in the
     packet (such as nested route errors) that need to be delivered */
  if (verbose_)
    trace("Sdebug %.5f _%s_ unwrapping nested route error",
          Scheduler::instance().clock(), net_id.dump());

  CBRP_Packet p_copy = p;
  p_copy.pkt = p.pkt->copy();

  hdr_cbrp *new_cbrph = (hdr_cbrp*)p_copy.pkt->access(off_cbrp_);
  hdr_ip *new_iph = (hdr_ip*)p_copy.pkt->access(off_ip_);

  // remove us from the list of errors
  new_cbrph->num_route_errors() -= 1;
  new_cbrph->route_request() = 0;
  //new_cbrph->route_error() = 1;

  // send the packet to the person listed in what's now the last entry
  p_copy.dest = ID(new_cbrph->down_links()[new_cbrph->num_route_errors()-1].tell_addr, ::IP);
  p_copy.src = net_id;

  new_iph->dst() = p_copy.dest.addr;
  new_iph->dport() = RT_PORT;
  new_iph->src() = p_copy.src.addr;
  new_iph->sport() = RT_PORT;
  new_iph->ttl() = 255;

  // an error packet is a first class citizen, so we'll
  // use handlePktWOSR to obtain a route if needed
  handlePktWithoutSR(p_copy, false);
}

/*==============================================================
  Callback for link layer transmission failures
------------------------------------------------------------*/
struct filterfailuredata {
  nsaddr_t dead_next_hop;
  int off_cmn_;
  CBRP_Agent *agent;
};

int
CBRP_FilterFailure(Packet *p, void *data)
{
  struct filterfailuredata *ffd = (filterfailuredata *) data;
  hdr_cmn *cmh = (hdr_cmn*)p->access(ffd->off_cmn_);
  int remove = cmh->next_hop() == ffd->dead_next_hop;
  if (remove) ffd->agent->undeliverablePkt(p,1);
  return remove;
}

void
CBRP_Agent::undeliverablePkt(Packet *pkt, int mine)
/* try our best to save the undeliverable packet using local repair or cache */
{
  hdr_cbrp *cbrph = (hdr_cbrp*)pkt->access(off_cbrp_);
  hdr_ip *iph = (hdr_ip*)pkt->access(off_ip_);
  hdr_cmn *cmh = (hdr_cmn*)pkt->access(off_cmn_);

  CBRP_Packet p(pkt,cbrph);
  p.dest = ID(iph->dst(),::IP);
  p.src = ID(iph->src(),::IP);
  p.pkt = mine ? pkt : pkt->copy();

  cbrph =  (hdr_cbrp*)p.pkt->access(off_cbrp_);
  iph = (hdr_ip*)p.pkt->access(off_ip_);
  cmh = (hdr_cmn*)p.pkt->access(off_cmn_);

  if ((cbrph->route_request()))

    {
       if (UnicastRREQ(p,iph->dst())) {
         trace("CBRP %.5f _%d_ salvage-request %d->%d", Scheduler::instance().clock(), myaddr_, iph->src(),cbrph->request_destination());
       }
       return;
     }
  else if (cbrph->route_reply())
    {
      // we have already dropped the gratuitous route reply packet
      assert(!cbrph->route_repaired());
      //try to salvage the route reply
#ifdef CBRP_DEBUG
      printf("cbrp %d %.9f trying to salvage route reply %d->%di\n",net_id.getNSAddr_t(),Scheduler::instance().clock(),iph->src(),iph->dst());
#endif

      if ((ntable->my_status == CLUSTER_HEAD) && (cbrph->cur_addr()<(cbrph->num_addrs()-1)) 
          && (cbrph->addrs[cbrph->cur_addr()+1].addr == myaddr_) ) {
        cbrph->cur_addr_++;
	p.route.setIterator(cbrph->cur_addr());
      }
      if (cbrph->reply_addrs()[cbrph->route_reply_len()-1].addr == net_id.getNSAddr_t()) {
        cbrph->route_reply_len()--;
      } 
        
      p.dest.addr = iph->dst();
      p.dest.type = IP;
      if (handleRREP(p)) {
        trace("CBRP %.9f _%d_ salvage-reply %d->%d",Scheduler::instance().clock(), myaddr_, iph->src(),iph->dst());
      }
      return;
    }
  else if (cbrph->route_shortening()) 
   {
      trace("CBRP %.5f _%d_ wrong-shorten %d...%d->%d...%d",Scheduler::instance().clock(),myaddr_,iph->src(),myaddr_,cbrph->addrs[cbrph->cur_addr()].addr,iph->dst());
      //My shortening of the path is wrong, don't do it
      cbrph->route_shortening() = 0;
      //reset the cur_addr_ pointer
      cbrph->cur_addr() = cbrph->cur_addr()-2;

      p.dest = ID(iph->dst(),::IP);
      p.src = ID(iph->src(),::IP);
      p.route.setIterator(cbrph->cur_addr());

      sendOutPacketWithRoute(p,false);

      //do not send any route error message
      return;
   }
 
  // it's a packet we're forwarding for someone, save it if we can...
  Path salvage_route;

  //firstly reset the cur_addr_ pointer correctly
  cbrph->cur_addr() -= 1;
  p.route.setIterator(cbrph->cur_addr());
  if (cbrph->cur_addr() >= (cbrph->num_addrs() - 1)) {
    trace("SDFU: route error beyond end of source route????");
    fprintf(stderr,"SDFU: route error beyond end of source route????\n");
    Packet::free(p.pkt);
    return;
  }

 //salvage the route with local two hop topology information
  if (cbrp_local_repair && ((cbrph->route_repaired() < MAX_LOCAL_REPAIR_TIMES ) && ((unsigned long)cbrph->addrs[cbrph->cur_addr()].addr == net_id.addr)) )
      { 
        if ((p.route.index()<p.route.length()-2) && 
            (cmh->next_hop() = ntable->GetQuickNextNode(p.route[p.route.index()+2].addr))) 
          {
             //try to reach the hop after next using another intermediate node
             trace("CBRP %.5f _%d_ salvage-repair link %d -> (%d) %d ", Scheduler::instance().clock(),myaddr_,myaddr_, cmh->next_hop(), cbrph->addrs[cbrph->cur_addr()+2].addr);
             cbrph->addrs[cbrph->cur_addr()+1].addr = cmh->next_hop();
             cbrph->addrs[cbrph->cur_addr()+1].addr_type = AF_INET;
             cbrph->cur_addr()++;
          }
         else if ((cmh->next_hop() = ntable->GetQuickNextNode((p.route[p.route.index()+1]).addr)))
          {
             //try to reach the next hop using a new intermediate node
             trace("CBRP %.5f _%d_ salvage-repair link %d -> (%d) %d ", Scheduler::instance().clock(),myaddr_,myaddr_, cmh->next_hop(), cbrph->addrs[cbrph->cur_addr()+1].addr);
             cbrph->cur_addr()++;
             for (int i = cbrph->num_addrs();i>cbrph->cur_addr();i--) {
               cbrph->addrs[i] = cbrph->addrs[i-1];
             }
             cbrph->num_addrs() += 1;
             cbrph->addrs[cbrph->cur_addr()].addr = cmh->next_hop();
             cbrph->addrs[cbrph->cur_addr()].addr_type = AF_INET;
           }
   
          if (cmh->next_hop()) {  //we've managed to save using local repair, send it out

             if (cbrph->cur_addr() == 1) {
               assert(p.src == net_id);
               //amend the route directly in the cache, do not need to mark repair flag
               Path route;
               //fill the path
               route.setLength(cbrph->num_addrs());
               for (int i=0;i<cbrph->num_addrs();i++) {
                  route[i] = ID(cbrph->addrs[i]);
               }   
               route.resetIterator();
               route_cache->addRoute(route, Scheduler::instance().clock(), p.dest);
             }else {
               cbrph->route_repaired()++; 
             }
             cmh->size() += cbrph->size();
             Scheduler::instance().schedule(ll,p.pkt, 0);
             return;
          }

      }

    if (cbrp_salvage_with_cache && route_cache->findRoute(p.dest, salvage_route, 0))
      { 
        p.route = salvage_route;
        p.route.setIterator(1);
        fillCBRPPath(p.route,cbrph);
        cmh->next_hop() = cbrph->addrs[p.route.index()].addr;
        trace("CBRP %.5f _%s_ salvage-cache %s -> %s --- %d with %s",
                Scheduler::instance().clock(), net_id.dump(),
                p.src.dump(), p.dest.dump(), cmh->uid(), p.route.dump());
        //prevent the route to be saved again in future.
        cbrph->route_repaired()=MAX_LOCAL_REPAIR_TIMES+1; 
        cmh->size() += cbrph->size();
        Scheduler::instance().schedule(ll,p.pkt, 0);
        return;
      }

     if (ID(iph->src(),::IP) == net_id)
      {
         // it's our packet we couldn't send, we do another RREQ
         trace("cbrp SAVVVE by re-handle it %d -> %d old src route %s",iph->src(),iph->dst(),p
.route.dump());
         cbrph->init();
         handlePktWithoutSR(p, false);
         return;
       }    

      //we failed in all previous attempts, no hope, say byebye to the packet
      trace("Ssalv %.5f _%s_ dropping --- %d %s -> %s %s CBR %d",
          Scheduler::instance().clock(),
          net_id.dump(), cmh->uid(), p.src.dump(), p.dest.dump(), p.route.dump(), cbrph->valid());
         if (mine) drop(pkt, DROP_RTR_NO_ROUTE);
      
     
    
}

#ifdef USE_GOD_FEEDBACK
static int linkerr_is_wrong = 0;
#endif

void
CBRP_Agent::xmitFailed(Packet *pkt)
/* this function is called when the link layer fails to send out a cbrp packet
   we will firstly use this error information to update neighbor table, adjacency table etc.
   then we will run our local repair algorithm to save data packets and RREP
   as well as sending out route error message to the source if it's data packet */
{
  hdr_cbrp *cbrph = (hdr_cbrp*)pkt->access(off_cbrp_);
  hdr_ip *iph = (hdr_ip*)pkt->access(off_ip_);
  hdr_cmn *cmh = (hdr_cmn*)pkt->access(off_cmn_);

  ID tell_id;
  ID from_id;
  ID to_id;

  assert(cmh->size() >= 0);

  if (verbose_)
    trace("SSendFailure %.9f _%s_ --- %d - %s, (to %d) RREQ %d(%d,%d), RREP %d",
      Scheduler::instance().clock(), net_id.dump(), cmh->uid(), cbrph->dump(),cmh->next_hop(), cbrph->request_destination(),iph->src(),cbrph->rtreq_seq(),cbrph->route_reply());


  //Jinyang -first update neighbor table to correct not-so-fast timeout event leaveouts
  //again - we did not take into consideration of the uni-directional links
  ntable_ent *ent;
  if ((ent = ntable->GetEntry(cmh->next_hop()))) {
#ifdef CBRP_DEBUG
    trace("cbrp _%d_ delete %d from neighbor table last update %.9f",myaddr_,cmh->next_hop(),ent->last_update);
#endif
    ntable->DeleteEntry(cmh->next_hop());
  }

  // if it's a gratuitous route reply about a repaired route, just drop it silently
  if (cbrph->route_reply() && (cbrph->route_repaired()||cbrph->route_shortened())) {
    Packet::free(pkt);
    return;
  }

  cmh->size() -= cbrph->size();  //recalcuate the size  

  from_id = net_id;
  to_id.addr = cmh->next_hop();
  to_id.type = (ID_Type)AF_INET;

  if (!cbrph->route_request() && !cbrph->route_reply() && !cbrph->route_repaired()) {
    //if this is a unrepaired data packet, we have to inform the packet src about the error
    //the packet src may not be the packet originator
    tell_id.addr = cbrph->addrs[0].addr;
    tell_id.type = (ID_Type) AF_INET;
  }
  
#ifdef USE_GOD_FEEDBACK
  if (God::instance()->hops(from_id.getNSAddr_t(), to_id.getNSAddr_t()) == 1)
    { /* god thinks this link is still valid */
      linkerr_is_wrong++;
      trace("SxmitFailed %.5f _%s_  %d->%d god okays #%d",
            Scheduler::instance().clock(), net_id.dump(),
            from_id.getNSAddr_t(), to_id.getNSAddr_t(), linkerr_is_wrong);
      fprintf(stderr,
              "xmitFailed on link %d->%d god okays - ignoring & recycling #%d\n"
,
              from_id.getNSAddr_t(), to_id.getNSAddr_t(), linkerr_is_wrong);
      /* put packet back on end of ifq for xmission */
      // make sure we aren't cycling packets
      assert(p.pkt->incoming == 0); // this is an outgoing packet
      ll->recv(pkt, (Handler*) 0);
      return;
    }
#endif

  /* kill any routes we have using this link */
  route_cache->noticeDeadLink(from_id, to_id,
                              Scheduler::instance().clock());

  /* give ourselves a chance to save the packet, only nodes on the original route save the pkt, local repair nodes do not */
  undeliverablePkt(pkt->copy(), 1);
  //if (Scheduler::instance().clock()>78.9) printf("2\n");

  /* now kill all the other packets in the output queue that would
     use the same next hop.  This is reasonable, since 802.11 has
     already retried the xmission multiple times => a persistent failure. */
  Packet *r, *nr, *head = 0;            // pkts to be recycled
  while((r = ifq->filter(to_id.getNSAddr_t()))) {
    r->next_ = head;
    head = r;
  }

  hdr_cbrp *cbrph_tmp;

  for(r = head; r; r = nr) {
    nr = r->next_;
    cbrph_tmp = (hdr_cbrp*)r->access(off_cbrp_);
    cmh->size() -= cbrph_tmp->size();
    // if it's a gratuitous route reply about a repaired route, just drop it silently
    if (cbrph_tmp->route_reply() && (cbrph_tmp->route_repaired()||cbrph_tmp->route_shortened())) {
      Packet::free(pkt);
      return;
    }       
    undeliverablePkt(r, 1);
  }

  /* warp pkt into a route error message */
  if (tell_id == net_id || tell_id == MAC_id )
    { // no need to send the route error if it's for us
      if (verbose_)
        trace("Sdebug _%s_ not bothering to send route error to myself",
              tell_id.dump());
      Packet::free(pkt);        
      pkt = 0;
      return;
    }
 
  if (cbrph->route_request() || (cbrph->route_reply()) || cbrph->route_repaired() || cbrph->route_shortening()) {
    Packet::free(pkt);
    pkt = 0;
    return;
  }
 
  if (cbrph->num_route_errors() >= MAX_ROUTE_ERRORS)
    { // no more room in the error packet to nest an additional error.
      // this pkt's been bouncing around so much, let's just drop and let
      // the originator retry
      // Another possibility is to just strip off the outer error, and
      // launch a Route discovey for the inner error XXX -dam 6/5/98
      trace("SDFU  %.5f _%s_ dumping maximally nested error %s  %d -> %d",
            Scheduler::instance().clock(), net_id.dump(),
            tell_id.dump(),
            from_id.dump(),
            to_id.dump());
      Packet::free(pkt);        // no drop needed
      pkt = 0;
      return;
    }

  cbrph->cur_addr()--;

  link_down_cbrp *deadlink = &(cbrph->down_links()[cbrph->num_route_errors()]);
  deadlink->addr_type = cbrph->addrs[cbrph->cur_addr()].addr_type;
  deadlink->from_addr = cbrph->addrs[cbrph->cur_addr()].addr;
  deadlink->to_addr = cbrph->addrs[cbrph->cur_addr()+1].addr;
  deadlink->tell_addr = cbrph->addrs[0].addr;
  cbrph->num_route_errors() += 1;

  if (verbose_)
    trace("Sdebug %.5f _%s_ sending into dead-link (nest %d) tell %d  %d -> %d",
          Scheduler::instance().clock(), net_id.dump(),
          cbrph->num_route_errors(),
          deadlink->tell_addr,
          deadlink->from_addr,
          deadlink->to_addr);

  cbrph->valid() = 1;

  cbrph->route_error() = 1;
  cbrph->route_reply() = 0;
  cbrph->route_request() = 0;

  iph->dst() = deadlink->tell_addr;
  iph->dport() = RT_PORT;
  iph->src() = net_id.addr;
  iph->sport() = RT_PORT;
  iph->ttl() = 255;

  cmh->ptype() = PT_CBRP;                
  cmh->num_forwards() = 0;
  cmh->size() = IP_HDR_LEN; //cut off data in the original packet
  // assign this packet a new uid, since we're sending it
  cmh->uid() = uidcnt_++;
  cbrph->num_addrs() = cbrph->cur_addr()+1;

  CBRP_Packet p(pkt, cbrph);
  p.route.setLength(p.route.index()+1);
  p.route.reverseInPlace();
  p.route.resetIterator();

  fillCBRPPath(p.route,cbrph);

  p.dest = tell_id;
  p.src = net_id;

  /* send out the Route Error message */
  sendOutPacketWithRoute(p, true);
}

void
CBRP_XmitFailureCallback(Packet *pkt, void *data)
{
  CBRP_Agent *agent = (CBRP_Agent *)data; // cast of trust
  agent->xmitFailed(pkt);
}
/* fill the path information in CBRP header */
void CBRP_Agent::fillCBRPPath(Path &path, hdr_cbrp *& cbrph)
{
  
  for (int i=0;i<path.length();i++) {
    path[i].fillSRAddr(cbrph->addrs[i]);
  }  
  cbrph->num_addrs() = path.length();
  cbrph->cur_addr() =  path.index();
}
